from scipy import sparse as sp
from discretize.utils import (
    sub2ind,
    sdiag,
    inverse_property_tensor,
    TensorType,
    make_property_tensor,
    ndgrid,
    inverse_2x2_block_diagonal,
    get_subarray,
    inverse_3x3_block_diagonal,
    spzeros,
    sdinv,
)
import numpy as np
from discretize.utils.code_utils import deprecate_method
import warnings


class InnerProducts(object):
    """Class for constructing inner product matrices.

    ``InnerProducts`` is a mixin class for constructing inner product matrices,
    their inverses and their derivatives with respect to model parameters.
    The ``InnerProducts`` class is inherited by all ``discretize`` mesh classes.
    In practice, we don't create instances of the ``InnerProducts`` class in
    order to construct inner product matrices, their inverses or their derivatives.
    These quantities are instead constructed from instances of ``discretize``
    meshes using the appropriate method.
    """

    def get_face_inner_product(
        self,
        model=None,
        invert_model=False,
        invert_matrix=False,
        do_fast=True,
        **kwargs
    ):
        r"""Generate the face inner product matrix or its inverse.

        This method generates the inner product matrix (or its inverse)
        when discrete variables are defined on mesh faces. It is also capable of
        constructing the inner product matrix when physical properties
        are defined in the form of constitutive relations. For a comprehensive
        description of the inner product matrices that can be constructed
        with **get_face_inner_product**, see *Notes*.

        Parameters
        ----------
        model : None or numpy.ndarray, optional
            Parameters defining the material properties for every cell in the mesh.
            Inner product matrices can be constructed for the following cases:

            - *None* : returns the basic inner product matrix
            - *(n_cells)* :class:`numpy.ndarray` : returns inner product matrix for an
              isotropic model. The array contains a scalar physical property value for
              each cell.
            - *(n_cells, dim)* :class:`numpy.ndarray` : returns inner product matrix for
              diagonal anisotropic case. Columns are ordered ``np.c_[σ_xx, σ_yy, σ_zz]``.
              This can also a be a 1D array with the same number of total elements in
              column major order.
            - *(n_cells, 3)* :class:`numpy.ndarray` (``dim`` is 2) or
              *(n_cells, 6)* :class:`numpy.ndarray` (``dim`` is 3) :
              returns inner product matrix for full tensor properties case. Columns are
              ordered ``np.c_[σ_xx, σ_yy, σ_zz, σ_xy, σ_xz, σ_yz]`` This can also be a
              1D array with the same number of total elements in column major order.

        invert_model : bool, optional
            The inverse of *model* is used as the physical property.
        invert_matrix : bool, optional
            Returns the inverse of the inner product matrix.
            The inverse not implemented for full tensor properties.
        do_fast : bool, optional
            Do a faster implementation (if available).

        Returns
        -------
        (n_faces, n_faces) scipy.sparse.csr_matrix
            inner product matrix

        Notes
        -----
        For continuous vector quantities :math:`\vec{u}` and :math:`\vec{w}`
        whose discrete representations :math:`\mathbf{u}` and :math:`\mathbf{w}`
        live on the faces, **get_face_inner_product** constructs the inner product matrix
        :math:`\mathbf{M_\ast}` (or its inverse :math:`\mathbf{M_\ast^{-1}}`) for the
        following cases:

        **Basic Inner Product:** the inner product between :math:`\vec{u}` and :math:`\vec{w}`

        .. math::
            \langle \vec{u}, \vec{w} \rangle = \mathbf{u^T \, M \, w}

        **Isotropic Case:** the inner product between :math:`\vec{u}` and :math:`\sigma \vec{w}`
        where :math:`\sigma` is a scalar function.

        .. math::
            \langle \vec{u}, \sigma \vec{w} \rangle = \mathbf{u^T \, M_\sigma \, w}

        **Tensor Case:** the inner product between :math:`\vec{u}` and :math:`\Sigma \vec{w}`
        where :math:`\Sigma` is tensor function; :math:`\sigma_{xy} = \sigma_{xz} = \sigma_{yz} = 0`
        for diagonal anisotropy.

        .. math::
            \langle \vec{u}, \Sigma \vec{w} \rangle = \mathbf{u^T \, M_\Sigma \, w}
            \;\;\; \textrm{where} \;\;\;
            \Sigma = \begin{bmatrix}
            \sigma_{xx} & \sigma_{xy} & \sigma_{xz} \\
            \sigma_{xy} & \sigma_{yy} & \sigma_{yz} \\
            \sigma_{xz} & \sigma_{yz} & \sigma_{zz}
            \end{bmatrix}

        Examples
        --------
        Here we provide some examples of face inner product matrices.
        For simplicity, we will work on a 2 x 2 x 2 tensor mesh.
        As seen below, we begin by constructing and imaging the basic
        face inner product matrix.

        >>> from discretize import TensorMesh
        >>> import matplotlib.pyplot as plt
        >>> import numpy as np
        >>> import matplotlib as mpl

        >>> h = np.ones(2)
        >>> mesh = TensorMesh([h, h, h])
        >>> Mf = mesh.get_face_inner_product()

        >>> fig = plt.figure(figsize=(6, 6))
        >>> ax = fig.add_subplot(111)
        >>> ax.imshow(Mf.todense())
        >>> ax.set_title('Basic Face Inner Product Matrix', fontsize=18)
        >>> plt.show()

        Next, we consider the case where the physical properties
        of the cells are defined by consistutive relations. For
        the isotropic, diagonal anisotropic and full tensor cases,
        we show the physical property tensor for a single cell.

        Define 4 constitutive parameters and define the tensor
        for each cell for isotropic, diagonal and tensor cases.

        >>> sig1, sig2, sig3, sig4, sig5, sig6 = 6, 5, 4, 3, 2, 1
        >>> sig_iso_tensor = sig1 * np.eye(3)
        >>> sig_diag_tensor = np.diag(np.array([sig1, sig2, sig3]))
        >>> sig_full_tensor = np.array([
        ...     [sig1, sig4, sig5],
        ...     [sig4, sig2, sig6],
        ...     [sig5, sig6, sig3]
        ... ])

        Then plot matrix entries,

        .. collapse:: Expand to show scripting for plot

            >>> fig = plt.figure(figsize=(15, 5))
            >>> ax1 = fig.add_subplot(131)
            >>> ax1.imshow(sig_iso_tensor)
            >>> ax1.axis('off')
            >>> ax1.set_title("Tensor (isotropic)", fontsize=16)
            >>> ax2 = fig.add_subplot(132)
            >>> ax2.imshow(sig_diag_tensor)
            >>> ax2.axis('off')
            >>> ax2.set_title("Tensor (diagonal anisotropic)", fontsize=16)
            >>> ax3 = fig.add_subplot(133)
            >>> ax3.imshow(sig_full_tensor)
            >>> ax3.axis('off')
            >>> ax3.set_title("Tensor (full anisotropic)", fontsize=16)
            >>> plt.show()

        Here, construct and image the face inner product matrices for
        the isotropic, diagonal anisotropic and full tensor cases.
        Spy plots are used to demonstrate the sparsity of the inner
        product matrices.

        Isotropic case:

        >>> v = np.ones(mesh.nC)
        >>> sig = sig1 * v
        >>> M1 = mesh.get_face_inner_product(sig)

        Diagonal anisotropic case:

        >>> sig = np.c_[sig1*v, sig2*v, sig3*v]
        >>> M2 = mesh.get_face_inner_product(sig)

        Full anisotropic case:

        >>> sig = np.tile(np.c_[sig1, sig2, sig3, sig4, sig5, sig6], (mesh.nC, 1))
        >>> M3 = mesh.get_face_inner_product(sig)

        And then we can plot the sparse representation,

        .. collapse:: Expand to show scripting for plot

            >>> fig = plt.figure(figsize=(12, 4))
            >>> ax1 = fig.add_subplot(131)
            >>> ax1.spy(M1, ms=5)
            >>> ax1.set_title("M (isotropic)", fontsize=16)
            >>> ax2 = fig.add_subplot(132)
            >>> ax2.spy(M2, ms=5)
            >>> ax2.set_title("M (diagonal anisotropic)", fontsize=16)
            >>> ax3 = fig.add_subplot(133)
            >>> ax3.spy(M3, ms=5)
            >>> ax3.set_title("M (full anisotropic)", fontsize=16)
            >>> plt.show()
        """
        if "invProp" in kwargs:
            warnings.warn(
                "The invProp keyword argument has been deprecated, please use invert_model. "
                "This will be removed in discretize 1.0.0",
                DeprecationWarning,
            )
            invert_model = kwargs["invProp"]
        if "invMat" in kwargs:
            warnings.warn(
                "The invMat keyword argument has been deprecated, please use invert_matrix. "
                "This will be removed in discretize 1.0.0",
                DeprecationWarning,
            )
            invert_matrix = kwargs["invMat"]
        if "doFast" in kwargs:
            warnings.warn(
                "The doFast keyword argument has been deprecated, please use do_fast. "
                "This will be removed in discretize 1.0.0",
                DeprecationWarning,
            )
            do_fast = kwargs["doFast"]

        return self._getInnerProduct(
            "F",
            model=model,
            invert_model=invert_model,
            invert_matrix=invert_matrix,
            do_fast=do_fast,
        )

    def get_edge_inner_product(
        self,
        model=None,
        invert_model=False,
        invert_matrix=False,
        do_fast=True,
        **kwargs
    ):
        r"""Generate the edge inner product matrix or its inverse.

        This method generates the inner product matrix (or its inverse)
        when discrete variables are defined on mesh edges. It is also capable of
        constructing the inner product matrix when physical properties
        are defined in the form of constitutive relations. For a comprehensive
        description of the inner product matrices that can be constructed
        with **get_edge_inner_product**, see *Notes*.

        Parameters
        ----------
        model : None or numpy.ndarray
            Parameters defining the material properties for every cell in the mesh.
            Inner product matrices can be constructed for the following cases:

            - *None* : returns the basic inner product matrix
            - *(n_cells)* :class:`numpy.ndarray` : returns inner product matrix for an
              isotropic model. The array contains a scalar physical property value for
              each cell.
            - *(n_cells, dim)* :class:`numpy.ndarray` : returns inner product matrix for
              diagonal anisotropic case. Columns are ordered ``np.c_[σ_xx, σ_yy, σ_zz]``.
              This can also a be a 1D array with the same number of total elements in
              column major order.
            - *(n_cells, 3)* :class:`numpy.ndarray` (``dim`` is 2) or
              *(n_cells, 6)* :class:`numpy.ndarray` (``dim`` is 3) :
              returns inner product matrix for full tensor properties case. Columns are
              ordered ``np.c_[σ_xx, σ_yy, σ_zz, σ_xy, σ_xz, σ_yz]`` This can also be a
              1D array with the same number of total elements in column major order.

        invert_model : bool, optional
            The inverse of *model* is used as the physical property.
        invert_matrix : bool, optional
            Teturns the inverse of the inner product matrix.
            The inverse not implemented for full tensor properties.
        do_fast : bool, optional
            Do a faster implementation (if available).

        Returns
        -------
        (n_edges, n_edges) scipy.sparse.csr_matrix
            inner product matrix

        Notes
        -----
        For continuous vector quantities :math:`\vec{u}` and :math:`\vec{w}`
        whose discrete representations :math:`\mathbf{u}` and :math:`\mathbf{w}`
        live on the edges, **get_edge_inner_product** constructs the inner product
        matrix :math:`\mathbf{M_\ast}` (or its inverse :math:`\mathbf{M_\ast^{-1}}`) for
        the following cases:

        **Basic Inner Product:** the inner product between :math:`\vec{u}` and
        :math:`\vec{w}`.

        .. math::
            \langle \vec{u}, \vec{w} \rangle = \mathbf{u^T \, M \, w}

        **Isotropic Case:** the inner product between :math:`\vec{u}` and
        :math:`\sigma \vec{w}` where :math:`\sigma` is a scalar function.

        .. math::
            \langle \vec{u}, \sigma \vec{w} \rangle = \mathbf{u^T \, M_\sigma \, w}

        **Tensor Case:** the inner product between :math:`\vec{u}` and
        :math:`\Sigma \vec{w}` where :math:`\Sigma` is tensor function;
        :math:`\sigma_{xy} = \sigma_{xz} = \sigma_{yz} = 0` for diagonal anisotropy.

        .. math::
            \langle \vec{u}, \Sigma \vec{w} \rangle =
            \mathbf{u^T \, M_\Sigma \, w} \;\;\; \textrm{where} \;\;\;
            \Sigma = \begin{bmatrix}
            \sigma_{xx} & \sigma_{xy} & \sigma_{xz} \\
            \sigma_{xy} & \sigma_{yy} & \sigma_{yz} \\
            \sigma_{xz} & \sigma_{yz} & \sigma_{zz}
            \end{bmatrix}

        Examples
        --------
        Here we provide some examples of edge inner product matrices.
        For simplicity, we will work on a 2 x 2 x 2 tensor mesh.
        As seen below, we begin by constructing and imaging the basic
        edge inner product matrix.

        >>> from discretize import TensorMesh
        >>> import matplotlib.pyplot as plt
        >>> import numpy as np
        >>> import matplotlib as mpl

        >>> h = np.ones(2)
        >>> mesh = TensorMesh([h, h, h])
        >>> Me = mesh.get_edge_inner_product()

        >>> fig = plt.figure(figsize=(6, 6))
        >>> ax = fig.add_subplot(111)
        >>> ax.imshow(Me.todense())
        >>> ax.set_title('Basic Edge Inner Product Matrix', fontsize=18)
        >>> plt.show()

        Next, we consider the case where the physical properties
        of the cells are defined by consistutive relations. For
        the isotropic, diagonal anisotropic and full tensor cases,
        we show the physical property tensor for a single cell.

        Define 4 constitutive parameters and define the tensor
        for each cell for isotropic, diagonal and tensor cases.

        >>> sig1, sig2, sig3, sig4, sig5, sig6 = 6, 5, 4, 3, 2, 1
        >>> sig_iso_tensor = sig1 * np.eye(3)
        >>> sig_diag_tensor = np.diag(np.array([sig1, sig2, sig3]))
        >>> sig_full_tensor = np.array([
        ...     [sig1, sig4, sig5],
        ...     [sig4, sig2, sig6],
        ...     [sig5, sig6, sig3]
        ... ])

        Then plot the matrix entries,

        .. collapse:: Expand to show scripting for plot

            >>> fig = plt.figure(figsize=(15, 5))
            >>> ax1 = fig.add_subplot(131)
            >>> ax1.imshow(sig_iso_tensor)
            >>> ax1.axis('off')
            >>> ax1.set_title("Tensor (isotropic)", fontsize=16)
            >>> ax2 = fig.add_subplot(132)
            >>> ax2.imshow(sig_diag_tensor)
            >>> ax2.axis('off')
            >>> ax2.set_title("Tensor (diagonal anisotropic)", fontsize=16)
            >>> ax3 = fig.add_subplot(133)
            >>> ax3.imshow(sig_full_tensor)
            >>> ax3.axis('off')
            >>> ax3.set_title("Tensor (full anisotropic)", fontsize=16)
            >>> plt.show()

        Here construct and image the edge inner product matrices for
        the isotropic, diagonal anisotropic and full tensor cases.
        Spy plots are used to demonstrate the sparsity of the inner
        product matrices.

        Isotropic case:

        >>> v = np.ones(mesh.nC)
        >>> sig = sig1 * v
        >>> M1 = mesh.get_edge_inner_product(sig)

        Diagonal anisotropic case:

        >>> sig = np.c_[sig1*v, sig2*v, sig3*v]
        >>> M2 = mesh.get_edge_inner_product(sig)

        Full anisotropic

        >>> sig = np.tile(np.c_[sig1, sig2, sig3, sig4, sig5, sig6], (mesh.nC, 1))
        >>> M3 = mesh.get_edge_inner_product(sig)

        Then plot the sparse representation,

        .. collapse:: Expand to show scripting for plot

            >>> fig = plt.figure(figsize=(12, 4))
            >>> ax1 = fig.add_subplot(131)
            >>> ax1.spy(M1, ms=5)
            >>> ax1.set_title("M (isotropic)", fontsize=16)
            >>> ax2 = fig.add_subplot(132)
            >>> ax2.spy(M2, ms=5)
            >>> ax2.set_title("M (diagonal anisotropic)", fontsize=16)
            >>> ax3 = fig.add_subplot(133)
            >>> ax3.spy(M3, ms=5)
            >>> ax3.set_title("M (full anisotropic)", fontsize=16)
            >>> plt.show()

        """
        if "invProp" in kwargs:
            warnings.warn(
                "The invProp keyword argument has been deprecated, please use invert_model. "
                "This will be removed in discretize 1.0.0",
                DeprecationWarning,
            )
            invert_model = kwargs["invProp"]
        if "invMat" in kwargs:
            warnings.warn(
                "The invMat keyword argument has been deprecated, please use invert_matrix. "
                "This will be removed in discretize 1.0.0",
                DeprecationWarning,
            )
            invert_matrix = kwargs["invMat"]
        if "doFast" in kwargs:
            warnings.warn(
                "The doFast keyword argument has been deprecated, please use do_fast. "
                "This will be removed in discretize 1.0.0",
                DeprecationWarning,
            )
            do_fast = kwargs["doFast"]
        return self._getInnerProduct(
            "E",
            model=model,
            invert_model=invert_model,
            invert_matrix=invert_matrix,
            do_fast=do_fast,
        )

    def _getInnerProduct(
        self,
        projection_type,
        model=None,
        invert_model=False,
        invert_matrix=False,
        do_fast=True,
        **kwargs
    ):
        """get the inner product matrix

        Parameters
        ----------

        str : projection_type
            'F' for faces 'E' for edges

        numpy.ndarray : model
            material property (tensor properties are possible) at each cell center (nC, (1, 3, or 6))

        bool : invert_model
            inverts the material property

        bool : invert_matrix
            inverts the matrix

        bool : do_fast
            do a faster implementation if available.


        Returns
        -------

        scipy.sparse.csr_matrix
            M, the inner product matrix (nE, nE)

        """
        if "invProp" in kwargs:
            warnings.warn(
                "The invProp keyword argument has been deprecated, please use invert_model. "
                "This will be removed in discretize 1.0.0",
                DeprecationWarning,
            )
            invert_model = kwargs["invProp"]
        if "invMat" in kwargs:
            warnings.warn(
                "The invMat keyword argument has been deprecated, please use invert_matrix. "
                "This will be removed in discretize 1.0.0",
                DeprecationWarning,
            )
            invert_matrix = kwargs["invMat"]
        if "doFast" in kwargs:
            warnings.warn(
                "The doFast keyword argument has been deprecated, please use do_fast. "
                "This will be removed in discretize 1.0.0",
                DeprecationWarning,
            )
            do_fast = kwargs["doFast"]
        if projection_type not in ["F", "E"]:
            raise TypeError("projection_type must be 'F' for faces or 'E' for edges")

        fast = None
        if hasattr(self, "_fastInnerProduct") and do_fast:
            fast = self._fastInnerProduct(
                projection_type,
                model=model,
                invert_model=invert_model,
                invert_matrix=invert_matrix,
            )
        if fast is not None:
            return fast

        if invert_model:
            model = inverse_property_tensor(self, model)

        tensorType = TensorType(self, model)

        Mu = make_property_tensor(self, model)
        Ps = self._getInnerProductProjectionMatrices(projection_type, tensorType)
        A = np.sum([P.T * Mu * P for P in Ps])

        if invert_matrix and tensorType < 3:
            A = sdinv(A)
        elif invert_matrix and tensorType == 3:
            raise Exception("Solver needed to invert A.")

        return A

    def _getInnerProductProjectionMatrices(self, projection_type, tensorType):
        """
        Parameters
        ----------
        projection_type : str
            'F' for faces 'E' for edges

        tensorType : TensorType
            type of the tensor: TensorType(mesh, sigma)
        """
        if not isinstance(tensorType, TensorType):
            raise TypeError("tensorType must be an instance of TensorType.")
        if projection_type not in ["F", "E"]:
            raise TypeError("projection_type must be 'F' for faces or 'E' for edges")

        d = self.dim
        # We will multiply by sqrt on each side to keep symmetry
        V = sp.kron(sp.identity(d), sdiag(np.sqrt((2 ** (-d)) * self.cell_volumes)))

        nodes = ["000", "100", "010", "110", "001", "101", "011", "111"][: 2 ** d]

        if projection_type == "F":
            locs = {
                "000": [("fXm",), ("fXm", "fYm"), ("fXm", "fYm", "fZm")],
                "100": [("fXp",), ("fXp", "fYm"), ("fXp", "fYm", "fZm")],
                "010": [None, ("fXm", "fYp"), ("fXm", "fYp", "fZm")],
                "110": [None, ("fXp", "fYp"), ("fXp", "fYp", "fZm")],
                "001": [None, None, ("fXm", "fYm", "fZp")],
                "101": [None, None, ("fXp", "fYm", "fZp")],
                "011": [None, None, ("fXm", "fYp", "fZp")],
                "111": [None, None, ("fXp", "fYp", "fZp")],
            }
            proj = getattr(self, "_getFaceP" + ("x" * d))()

        elif projection_type == "E":
            locs = {
                "000": [("eX0",), ("eX0", "eY0"), ("eX0", "eY0", "eZ0")],
                "100": [("eX0",), ("eX0", "eY1"), ("eX0", "eY1", "eZ1")],
                "010": [None, ("eX1", "eY0"), ("eX1", "eY0", "eZ2")],
                "110": [None, ("eX1", "eY1"), ("eX1", "eY1", "eZ3")],
                "001": [None, None, ("eX2", "eY2", "eZ0")],
                "101": [None, None, ("eX2", "eY3", "eZ1")],
                "011": [None, None, ("eX3", "eY2", "eZ2")],
                "111": [None, None, ("eX3", "eY3", "eZ3")],
            }
            proj = getattr(self, "_getEdgeP" + ("x" * d))()

        return [V * proj(*locs[node][d - 1]) for node in nodes]

    def get_face_inner_product_deriv(
        self, model, do_fast=True, invert_model=False, invert_matrix=False, **kwargs
    ):
        r"""Function handle to multiply vector with derivative of face inner product matrix (or its inverse).

        Let :math:`\mathbf{M}(\mathbf{m})` be the face inner product matrix
        constructed with a set of physical property parameters :math:`\mathbf{m}`
        (or its inverse). **get_face_inner_product_deriv** constructs a function handle

        .. math::
            \mathbf{F}(\mathbf{u}) = \mathbf{u}^T \, \frac{\partial \mathbf{M}(\mathbf{m})}{\partial \mathbf{m}}

        which accepts any numpy.array :math:`\mathbf{u}` of shape (n_faces,). That is,
        **get_face_inner_product_deriv** constructs a function handle for computing
        the dot product between a vector :math:`\mathbf{u}` and the derivative of the
        face inner product matrix (or its inverse) with respect to the property parameters.
        When computed, :math:`\mathbf{F}(\mathbf{u})` returns a ``scipy.sparse.csr_matrix``
        of shape (n_faces, n_param).

        The function handle can be created for isotropic, diagonal
        isotropic and full tensor physical properties; see notes.

        Parameters
        ----------
        model : numpy.ndarray
            Parameters defining the material properties for every cell in the mesh.
            Inner product matrices can be constructed for the following cases:

            - *(n_cells)* :class:`numpy.ndarray` : Isotropic case. *model* contains a
              scalar physical property value for each cell.
            - *(n_cells, dim)* :class:`numpy.ndarray` : Diagonal anisotropic case.
              Columns are ordered ``np.c_[σ_xx, σ_yy, σ_zz]``. This can also a be a 1D
              array with the same number of total elements in column major order.
            - *(n_cells, 3)* :class:`numpy.ndarray` (``dim`` is 2) or
              *(n_cells, 6)* :class:`numpy.ndarray` (``dim`` is 3) : Full tensor properties case. Columns
              are ordered ``np.c_[σ_xx, σ_yy, σ_zz, σ_xy, σ_xz, σ_yz]`` This can also be
              a 1D array with the same number of total elements in column major order.

        invert_model : bool, optional
            The inverse of *model* is used as the physical property.
        invert_matrix : bool, optional
            Returns the inverse of the inner product matrix.
            The inverse not implemented for full tensor properties.
        do_fast : bool, optional
            Do a faster implementation (if available).

        Returns
        -------
        function
            The function handle :math:`\mathbf{F}(\mathbf{u})` which accepts a
            (``n_faces``) :class:`numpy.ndarray` :math:`\mathbf{u}`. The function
            returns a (``n_faces``, ``n_params``) :class:`scipy.sparse.csr_matrix`.

        Notes
        -----
        Let :math:`\mathbf{M}(\mathbf{m})` be the face inner product matrix (or its inverse)
        for the set of physical property parameters :math:`\mathbf{m}`. And let :math:`\mathbf{u}`
        be a discrete quantity that lives on the faces. **get_face_inner_product_deriv**
        creates a function handle for computing the following:

        .. math::
            \mathbf{F}(\mathbf{u}) = \mathbf{u}^T \, \frac{\partial \mathbf{M}(\mathbf{m})}{\partial \mathbf{m}}

        The dimensions of the sparse matrix constructed by computing :math:`\mathbf{F}(\mathbf{u})`
        for some :math:`\mathbf{u}` depends on the constitutive relation defined for each cell.
        These cases are summarized below.

        **Isotropic Case:** The physical property for each cell is defined by a scalar value.
        Therefore :math:`\mathbf{m}` is a (``n_cells``) :class:`numpy.ndarray`. The sparse matrix
        output by computing :math:`\mathbf{F}(\mathbf{u})` has shape (``n_faces``, ``n_cells``).

        **Diagonal Anisotropic Case:** In this case, the physical properties for each cell are
        defined by a diagonal tensor

        .. math::
            \Sigma = \begin{bmatrix}
            \sigma_{xx} & 0 & 0 \\
            0 & \sigma_{yy} & 0 \\
            0 & 0 & \sigma_{zz}
            \end{bmatrix}

        Thus there are ``dim * n_cells`` physical property parameters and :math:`\mathbf{m}` is
        a (``dim * n_cells``) :class:`numpy.ndarray`.  The sparse matrix
        output by computing :math:`\mathbf{F}(\mathbf{u})` has shape (``n_faces``, ``dim * n_cells``).

        **Full Tensor Case:** In this case, the physical properties for each cell are
        defined by a full tensor

        .. math::
            \Sigma = \begin{bmatrix}
            \sigma_{xx} & \sigma_{xy} & \sigma_{xz} \\
            \sigma_{xy} & \sigma_{yy} & \sigma_{yz} \\
            \sigma_{xz} & \sigma_{yz} & \sigma_{zz}
            \end{bmatrix}

        Thus there are ``6 * n_cells`` physical property parameters in 3 dimensions, or
        ``3 * n_cells`` physical property parameters in 2 dimensions, and
        :math:`\mathbf{m}` is a (``n_params``) :class:`numpy.ndarray`.
        The sparse matrix output by computing :math:`\mathbf{F}(\mathbf{u})`
        has shape (``n_faces``, ``n_params``).

        Examples
        --------
        Here, we construct a 4 cell by 4 cell tensor mesh. For our first example we
        consider isotropic physical properties; that is, the physical properties
        of each cell are defined a scalar value. We construct the face inner product
        matrix and visualize it with a spy plot. We then use
        **get_face_inner_product_deriv** to construct the function handle
        :math:`\mathbf{F}(\mathbf{u})` and plot the evaluation
        of this function on a spy plot.

        >>> from discretize import TensorMesh
        >>> import matplotlib.pyplot as plt
        >>> import numpy as np
        >>> import matplotlib as mpl
        >>> mpl.rcParams.update({'font.size': 14})
        >>> np.random.seed(45)
        >>> mesh = TensorMesh([[(1, 4)], [(1, 4)]])

        Define a model, and a random vector to multiply the derivative with,
        then we grab the respective derivative function and calculate the
        sparse matrix,

        >>> m = np.random.rand(mesh.nC)  # physical property parameters
        >>> u = np.random.rand(mesh.nF)  # vector of shape (n_faces)
        >>> Mf = mesh.get_face_inner_product(m)
        >>> F = mesh.get_face_inner_product_deriv(m)  # Function handle
        >>> dFdm_u = F(u)

        Spy plot for the inner product matrix and its derivative

        .. collapse:: Expand to show scripting for plot

            >>> fig = plt.figure(figsize=(15, 5))
            >>> ax1 = fig.add_axes([0.05, 0.05, 0.3, 0.85])
            >>> ax1.spy(Mf, ms=6)
            >>> ax1.set_title("Face Inner Product Matrix (Isotropic)", fontsize=14, pad=5)
            >>> ax1.set_xlabel("Face Index", fontsize=12)
            >>> ax1.set_ylabel("Face Index", fontsize=12)
            >>> ax2 = fig.add_axes([0.43, 0.05, 0.17, 0.8])
            >>> ax2.spy(dFdm_u, ms=6)
            >>> ax2.set_title(
            ...     "$u^T \, \dfrac{\partial M(m)}{\partial m}$ (Isotropic)",
            ...     fontsize=14, pad=5
            ... )
            >>> ax2.set_xlabel("Parameter Index", fontsize=12)
            >>> ax2.set_ylabel("Face Index", fontsize=12)
            >>> plt.show()

        For our second example, the physical properties on the mesh are fully
        anisotropic; that is, the physical properties of each cell are defined
        by a tensor with parameters :math:`\sigma_1`, :math:`\sigma_2` and :math:`\sigma_3`.
        Once again we construct the face inner product matrix and visualize it with a
        spy plot. We then use **get_face_inner_product_deriv** to construct the
        function handle :math:`\mathbf{F}(\mathbf{u})` and plot the evaluation
        of this function on a spy plot.

        >>> m = np.random.rand(mesh.nC, 3)  # anisotropic physical property parameters
        >>> u = np.random.rand(mesh.nF)     # vector of shape (n_faces)
        >>> Mf = mesh.get_face_inner_product(m)
        >>> F = mesh.get_face_inner_product_deriv(m)  # Function handle
        >>> dFdm_u = F(u)

        Plot the anisotropic inner product matrix and its derivative matrix,

        .. collapse:: Expand to show scripting for plot

            >>> fig = plt.figure(figsize=(15, 5))
            >>> ax1 = fig.add_axes([0.05, 0.05, 0.3, 0.8])
            >>> ax1.spy(Mf, ms=6)
            >>> ax1.set_title("Face Inner Product (Full Tensor)", fontsize=14, pad=5)
            >>> ax1.set_xlabel("Face Index", fontsize=12)
            >>> ax1.set_ylabel("Face Index", fontsize=12)
            >>> ax2 = fig.add_axes([0.4, 0.05, 0.45, 0.85])
            >>> ax2.spy(dFdm_u, ms=6)
            >>> ax2.set_title(
            ...     "$u^T \, \dfrac{\partial M(m)}{\partial m} \;$ (Full Tensor)",
            ...     fontsize=14, pad=5
            ... )
            >>> ax2.set_xlabel("Parameter Index", fontsize=12)
            >>> ax2.set_ylabel("Face Index", fontsize=12)
            >>> plt.show()

        """
        if "invProp" in kwargs:
            warnings.warn(
                "The invProp keyword argument has been deprecated, please use invert_model. "
                "This will be removed in discretize 1.0.0",
                DeprecationWarning,
            )
            invert_model = kwargs["invProp"]
        if "invMat" in kwargs:
            warnings.warn(
                "The invMat keyword argument has been deprecated, please use invert_matrix. "
                "This will be removed in discretize 1.0.0",
                DeprecationWarning,
            )
            invert_matrix = kwargs["invMat"]
        if "doFast" in kwargs:
            warnings.warn(
                "The doFast keyword argument has been deprecated, please use do_fast. "
                "This will be removed in discretize 1.0.0",
                DeprecationWarning,
            )
            do_fast = kwargs["doFast"]
        return self._getInnerProductDeriv(
            model,
            "F",
            do_fast=do_fast,
            invert_model=invert_model,
            invert_matrix=invert_matrix,
        )

    def get_edge_inner_product_deriv(
        self, model, do_fast=True, invert_model=False, invert_matrix=False, **kwargs
    ):
        r"""Function handle to multiply vector with derivative of edge inner product matrix (or its inverse).

        Let :math:`\mathbf{M}(\mathbf{m})` be the edge inner product matrix
        constructed with a set of physical property parameters :math:`\mathbf{m}`
        (or its inverse). **get_edge_inner_product_deriv** constructs a function handle

        .. math::
            \mathbf{F}(\mathbf{u}) = \mathbf{u}^T \, \frac{\partial \mathbf{M}(\mathbf{m})}{\partial \mathbf{m}}

        which accepts any numpy.array :math:`\mathbf{u}` of shape (n_edges,). That is,
        **get_edge_inner_product_deriv** constructs a function handle for computing
        the dot product between a vector :math:`\mathbf{u}` and the derivative of the
        edge inner product matrix (or its inverse) with respect to the property parameters.
        When computed, :math:`\mathbf{F}(\mathbf{u})` returns a ``scipy.sparse.csr_matrix``
        of shape (n_edges, n_param).

        The function handle can be created for isotropic, diagonal
        isotropic and full tensor physical properties; see notes.

        Parameters
        ----------
        model : numpy.ndarray
            Parameters defining the material properties for every cell in the mesh.

            Allows for the following cases:

            - *(n_cells)* :class:`numpy.ndarray` : Isotropic case. *model* contains a
              scalar physical property value for each cell.
            - *(n_cells, dim)* :class:`numpy.ndarray` : Diagonal anisotropic case.
              Columns are ordered ``np.c_[σ_xx, σ_yy, σ_zz]``. This can also a be a 1D
              array with the same number of total elements in column major order.
            - *(n_cells, 3)* :class:`numpy.ndarray` (``dim`` is 2) or
              *(n_cells, 6)* :class:`numpy.ndarray` (``dim`` is 3) : Full tensor properties case. Columns
              are ordered ``np.c_[σ_xx, σ_yy, σ_zz, σ_xy, σ_xz, σ_yz]`` This can also be
              a 1D array with the same number of total elements in column major order.

        invert_model : bool, optional
            The inverse of *model* is used as the physical property.
        invert_matrix : bool, optional
            Returns the function handle for the inverse of the inner product matrix
            The inverse not implemented for full tensor properties.
        do_fast : bool, optional
            Do a faster implementation (if available).

        Returns
        -------
        function
            The function handle :math:`\mathbf{F}(\mathbf{u})` which accepts a
            (``n_edges``) :class:`numpy.ndarray` :math:`\mathbf{u}`. The function
            returns a (``n_edges``, ``n_params``) :class:`scipy.sparse.csr_matrix`.

        Notes
        -----
        Let :math:`\mathbf{M}(\mathbf{m})` be the edge inner product matrix (or its inverse)
        for the set of physical property parameters :math:`\mathbf{m}`. And let :math:`\mathbf{u}`
        be a discrete quantity that lives on the edges. **get_edge_inner_product_deriv**
        creates a function handle for computing the following:

        .. math::
            \mathbf{F}(\mathbf{u}) = \mathbf{u}^T \, \frac{\partial \mathbf{M}(\mathbf{m})}{\partial \mathbf{m}}

        The dimensions of the sparse matrix constructed by computing :math:`\mathbf{F}(\mathbf{u})`
        for some :math:`\mathbf{u}` depends on the constitutive relation defined for each cell.
        These cases are summarized below.

        **Isotropic Case:** The physical property for each cell is defined by a scalar value.
        Therefore :math:`\mathbf{m}` is a (``n_cells``) :class:`numpy.ndarray`. The sparse matrix
        output by computing :math:`\mathbf{F}(\mathbf{u})` has shape (``n_edges``, ``n_cells``).

        **Diagonal Anisotropic Case:** In this case, the physical properties for each cell are
        defined by a diagonal tensor

        .. math::
            \Sigma = \begin{bmatrix}
            \sigma_{xx} & 0 & 0 \\
            0 & \sigma_{yy} & 0 \\
            0 & 0 & \sigma_{zz}
            \end{bmatrix}

        Thus there are ``dim * n_cells`` physical property parameters and :math:`\mathbf{m}` is
        a (``dim * n_cells``) :class:`numpy.ndarray`.  The sparse matrix
        output by computing :math:`\mathbf{F}(\mathbf{u})` has shape (``n_edges``, ``dim * n_cells``).

        **Full Tensor Case:** In this case, the physical properties for each cell are
        defined by a full tensor

        .. math::
            \Sigma = \begin{bmatrix}
            \sigma_{xx} & \sigma_{xy} & \sigma_{xz} \\
            \sigma_{xy} & \sigma_{yy} & \sigma_{yz} \\
            \sigma_{xz} & \sigma_{yz} & \sigma_{zz}
            \end{bmatrix}

        Thus there are ``6 * n_cells`` physical property parameters in 3 dimensions, or
        ``3 * n_cells`` physical property parameters in 2 dimensions, and :math:`\mathbf{m}` is
        a (``n_params``) :class:`numpy.ndarray`. The sparse matrix
        output by computing :math:`\mathbf{F}(\mathbf{u})` has shape (``n_edges``, ``n_params``).

        Examples
        --------
        Here, we construct a 4 cell by 4 cell tensor mesh. For our first example we
        consider isotropic physical properties; that is, the physical properties
        of each cell are defined a scalar value. We construct the edge inner product
        matrix and visualize it with a spy plot. We then use
        **get_edge_inner_product_deriv** to construct the function handle
        :math:`\mathbf{F}(\mathbf{u})` and plot the evaluation of this function on a spy
        plot.

        >>> from discretize import TensorMesh
        >>> import matplotlib.pyplot as plt
        >>> import numpy as np
        >>> import matplotlib as mpl
        >>> mpl.rcParams.update({'font.size': 14})
        >>> np.random.seed(45)
        >>> mesh = TensorMesh([[(1, 4)], [(1, 4)]])

        Next we create a random isotropic model vector, and a random vector to multiply
        the derivative with (for illustration purposes).

        >>> m = np.random.rand(mesh.nC)  # physical property parameters
        >>> u = np.random.rand(mesh.nF)  # vector of shape (n_edges)
        >>> Me = mesh.get_edge_inner_product(m)
        >>> F = mesh.get_edge_inner_product_deriv(m)  # Function handle
        >>> dFdm_u = F(u)

        Plot inner product matrix and its derivative matrix

        .. collapse:: Expand to show scripting for plot

            >>> fig = plt.figure(figsize=(15, 5))
            >>> ax1 = fig.add_axes([0.05, 0.05, 0.3, 0.8])
            >>> ax1.spy(Me, ms=6)
            >>> ax1.set_title("Edge Inner Product Matrix (Isotropic)", fontsize=14, pad=5)
            >>> ax1.set_xlabel("Edge Index", fontsize=12)
            >>> ax1.set_ylabel("Edge Index", fontsize=12)
            >>> ax2 = fig.add_axes([0.43, 0.05, 0.17, 0.8])
            >>> ax2.spy(dFdm_u, ms=6)
            >>> ax2.set_title(
            ...     "$u^T \, \dfrac{\partial M(m)}{\partial m}$ (Isotropic)",
            ...     fontsize=14, pad=5
            ... )
            >>> ax2.set_xlabel("Parameter Index", fontsize=12)
            >>> ax2.set_ylabel("Edge Index", fontsize=12)
            >>> plt.show()

        For our second example, the physical properties on the mesh are fully
        anisotropic; that is, the physical properties of each cell are defined
        by a tensor with parameters :math:`\sigma_1`, :math:`\sigma_2` and :math:`\sigma_3`.
        Once again we construct the edge inner product matrix and visualize it with a
        spy plot. We then use **get_edge_inner_product_deriv** to construct the
        function handle :math:`\mathbf{F}(\mathbf{u})` and plot the evaluation
        of this function on a spy plot.

        >>> m = np.random.rand(mesh.nC, 3)  # physical property parameters
        >>> u = np.random.rand(mesh.nF)     # vector of shape (n_edges)
        >>> Me = mesh.get_edge_inner_product(m)
        >>> F = mesh.get_edge_inner_product_deriv(m)  # Function handle
        >>> dFdm_u = F(u)

        Plot the anisotropic inner product matrix and its derivative matrix

        .. collapse:: Expand to show scripting for plot

            >>> fig = plt.figure(figsize=(15, 5))
            >>> ax1 = fig.add_axes([0.05, 0.05, 0.3, 0.8])
            >>> ax1.spy(Me, ms=6)
            >>> ax1.set_title("Edge Inner Product (Full Tensor)", fontsize=14, pad=5)
            >>> ax1.set_xlabel("Edge Index", fontsize=12)
            >>> ax1.set_ylabel("Edge Index", fontsize=12)
            >>> ax2 = fig.add_axes([0.4, 0.05, 0.45, 0.8])
            >>> ax2.spy(dFdm_u, ms=6)
            >>> ax2.set_title(
            ...     "$u^T \, \dfrac{\partial M(m)}{\partial m} \;$ (Full Tensor)",
            ...     fontsize=14, pad=5
            ... )
            >>> ax2.set_xlabel("Parameter Index", fontsize=12)
            >>> ax2.set_ylabel("Edge Index", fontsize=12)
            >>> plt.show()
        """
        if "invProp" in kwargs:
            warnings.warn(
                "The invProp keyword argument has been deprecated, please use invert_model. "
                "This will be removed in discretize 1.0.0",
                DeprecationWarning,
            )
            invert_model = kwargs["invProp"]
        if "invMat" in kwargs:
            warnings.warn(
                "The invMat keyword argument has been deprecated, please use invert_matrix. "
                "This will be removed in discretize 1.0.0",
                DeprecationWarning,
            )
            invert_matrix = kwargs["invMat"]
        if "doFast" in kwargs:
            warnings.warn(
                "The doFast keyword argument has been deprecated, please use do_fast. "
                "This will be removed in discretize 1.0.0",
                DeprecationWarning,
            )
            do_fast = kwargs["doFast"]
        return self._getInnerProductDeriv(
            model,
            "E",
            do_fast=do_fast,
            invert_model=invert_model,
            invert_matrix=invert_matrix,
        )

    def _getInnerProductDeriv(
        self,
        model,
        projection_type,
        do_fast=True,
        invert_model=False,
        invert_matrix=False,
    ):
        """
        Parameters
        ----------
        model : numpy.ndarray
            material property (tensor properties are possible) at each cell center (nC, (1, 3, or 6))

        projection_type : str
            'F' for faces 'E' for edges

        do_fast : bool
            do a faster implementation if available.

        invert_model : bool
            inverts the material property

        invert_matrix : bool
            inverts the matrix


        Returns
        -------
        scipy.sparse.csr_matrix
            dMdm, the derivative of the inner product matrix (nE, nC*nA)

        """
        fast = None
        if hasattr(self, "_fastInnerProductDeriv") and do_fast:
            fast = self._fastInnerProductDeriv(
                projection_type,
                model,
                invert_model=invert_model,
                invert_matrix=invert_matrix,
            )
        if fast is not None:
            return fast

        if invert_model or invert_matrix:
            raise NotImplementedError(
                "inverting the property or the matrix is not yet implemented for this mesh/tensorType. You should write it!"
            )

        tensorType = TensorType(self, model)
        P = self._getInnerProductProjectionMatrices(
            projection_type, tensorType=tensorType
        )

        def innerProductDeriv(v):
            return self._getInnerProductDerivFunction(tensorType, P, projection_type, v)

        return innerProductDeriv

    def _getInnerProductDerivFunction(self, tensorType, P, projection_type, v):
        """
        Parameters
        ----------
        model : numpy.ndarray
            material property (tensor properties are possible) at each cell center (nC, (1, 3, or 6))

        v : numpy.ndarray
            vector to multiply (required in the general implementation)

        P : list
            list of projection matrices

        projection_type : str
            'F' for faces 'E' for edges


        Returns
        -------
        scipy.sparse.csr_matrix
            dMdm, the derivative of the inner product matrix (n, nC*nA)

        """
        if projection_type not in ["F", "E"]:
            raise TypeError("projection_type must be 'F' for faces or 'E' for edges")

        n = getattr(self, "n" + projection_type)

        if tensorType == -1:
            return None

        if v is None:
            raise Exception("v must be supplied for this implementation.")

        d = self.dim
        Z = spzeros(self.nC, self.nC)

        if tensorType == 0:
            dMdm = spzeros(n, 1)
            for i, p in enumerate(P):
                dMdm = dMdm + sp.csr_matrix(
                    (p.T * (p * v), (range(n), np.zeros(n))), shape=(n, 1)
                )
        if d == 1:
            if tensorType == 1:
                dMdm = spzeros(n, self.nC)
                for i, p in enumerate(P):
                    dMdm = dMdm + p.T * sdiag(p * v)
        elif d == 2:
            if tensorType == 1:
                dMdm = spzeros(n, self.nC)
                for i, p in enumerate(P):
                    Y = p * v
                    y1 = Y[: self.nC]
                    y2 = Y[self.nC :]
                    dMdm = dMdm + p.T * sp.vstack((sdiag(y1), sdiag(y2)))
            elif tensorType == 2:
                dMdms = [spzeros(n, self.nC) for _ in range(2)]
                for i, p in enumerate(P):
                    Y = p * v
                    y1 = Y[: self.nC]
                    y2 = Y[self.nC :]
                    dMdms[0] = dMdms[0] + p.T * sp.vstack((sdiag(y1), Z))
                    dMdms[1] = dMdms[1] + p.T * sp.vstack((Z, sdiag(y2)))
                dMdm = sp.hstack(dMdms)
            elif tensorType == 3:
                dMdms = [spzeros(n, self.nC) for _ in range(3)]
                for i, p in enumerate(P):
                    Y = p * v
                    y1 = Y[: self.nC]
                    y2 = Y[self.nC :]
                    dMdms[0] = dMdms[0] + p.T * sp.vstack((sdiag(y1), Z))
                    dMdms[1] = dMdms[1] + p.T * sp.vstack((Z, sdiag(y2)))
                    dMdms[2] = dMdms[2] + p.T * sp.vstack((sdiag(y2), sdiag(y1)))
                dMdm = sp.hstack(dMdms)
        elif d == 3:
            if tensorType == 1:
                dMdm = spzeros(n, self.nC)
                for i, p in enumerate(P):
                    Y = p * v
                    y1 = Y[: self.nC]
                    y2 = Y[self.nC : self.nC * 2]
                    y3 = Y[self.nC * 2 :]
                    dMdm = dMdm + p.T * sp.vstack((sdiag(y1), sdiag(y2), sdiag(y3)))
            elif tensorType == 2:
                dMdms = [spzeros(n, self.nC) for _ in range(3)]
                for i, p in enumerate(P):
                    Y = p * v
                    y1 = Y[: self.nC]
                    y2 = Y[self.nC : self.nC * 2]
                    y3 = Y[self.nC * 2 :]
                    dMdms[0] = dMdms[0] + p.T * sp.vstack((sdiag(y1), Z, Z))
                    dMdms[1] = dMdms[1] + p.T * sp.vstack((Z, sdiag(y2), Z))
                    dMdms[2] = dMdms[2] + p.T * sp.vstack((Z, Z, sdiag(y3)))
                dMdm = sp.hstack(dMdms)
            elif tensorType == 3:
                dMdms = [spzeros(n, self.nC) for _ in range(6)]
                for i, p in enumerate(P):
                    Y = p * v
                    y1 = Y[: self.nC]
                    y2 = Y[self.nC : self.nC * 2]
                    y3 = Y[self.nC * 2 :]
                    dMdms[0] = dMdms[0] + p.T * sp.vstack((sdiag(y1), Z, Z))
                    dMdms[1] = dMdms[1] + p.T * sp.vstack((Z, sdiag(y2), Z))
                    dMdms[2] = dMdms[2] + p.T * sp.vstack((Z, Z, sdiag(y3)))
                    dMdms[3] = dMdms[3] + p.T * sp.vstack((sdiag(y2), sdiag(y1), Z))
                    dMdms[4] = dMdms[4] + p.T * sp.vstack((sdiag(y3), Z, sdiag(y1)))
                    dMdms[5] = dMdms[5] + p.T * sp.vstack((Z, sdiag(y3), sdiag(y2)))
                dMdm = sp.hstack(dMdms)

        return dMdm

    # ------------------------ Geometries ------------------------------
    #
    #
    #         node(i,j,k+1) ------ edge2(i,j,k+1) ----- node(i,j+1,k+1)
    #              /                                    /
    #             /                                    / |
    #         edge3(i,j,k)     face1(i,j,k)        edge3(i,j+1,k)
    #           /                                    /   |
    #          /                                    /    |
    #    node(i,j,k) ------ edge2(i,j,k) ----- node(i,j+1,k)
    #         |                                     |    |
    #         |                                     |   node(i+1,j+1,k+1)
    #         |                                     |    /
    #    edge1(i,j,k)      face3(i,j,k)        edge1(i,j+1,k)
    #         |                                     |  /
    #         |                                     | /
    #         |                                     |/
    #    node(i+1,j,k) ------ edge2(i+1,j,k) ----- node(i+1,j+1,k)

    def _getFacePx(M):
        """Returns a function for creating projection matrices"""
        ii = np.arange(M.shape_cells[0])

        def Px(xFace):
            """
            xFace is 'fXp' or 'fXm'
            """
            posFx = 0 if xFace == "fXm" else 1
            IND = ii + posFx
            PX = sp.csr_matrix((np.ones(M.nC), (range(M.nC), IND)), shape=(M.nC, M.nF))
            return PX

        return Px

    def _getFacePxx(M):
        """returns a function for creating projection matrices

        Mats takes you from faces a subset of all faces on only the
        faces that you ask for.

        These are centered around a single nodes.

        For example, if this was your entire mesh:

                        f3(Yp)
                  2_______________3
                  |               |
                  |               |
                  |               |
          f0(Xm)  |       x       |  f1(Xp)
                  |               |
                  |               |
                  |_______________|
                  0               1
                        f2(Ym)

        Pxx('fXm','fYm') = | 1, 0, 0, 0 |
                           | 0, 0, 1, 0 |

        Pxx('fXp','fYm') = | 0, 1, 0, 0 |
                           | 0, 0, 1, 0 |

        """
        i, j = np.arange(M.shape_cells[0]), np.arange(M.shape_cells[1])

        iijj = ndgrid(i, j)
        ii, jj = iijj[:, 0], iijj[:, 1]

        if M._meshType == "Curv":
            fN1 = M.reshape(M.face_normals, "F", "Fx", "M")
            fN2 = M.reshape(M.face_normals, "F", "Fy", "M")

        def Pxx(xFace, yFace):
            """
            xFace is 'fXp' or 'fXm'
            yFace is 'fYp' or 'fYm'
            """
            # no | node      | f1     | f2
            # 00 | i  ,j     | i  , j | i, j
            # 10 | i+1,j     | i+1, j | i, j
            # 01 | i  ,j+1   | i  , j | i, j+1
            # 11 | i+1,j+1   | i+1, j | i, j+1

            posFx = 0 if xFace == "fXm" else 1
            posFy = 0 if yFace == "fYm" else 1

            ind1 = sub2ind(M.vnFx, np.c_[ii + posFx, jj])
            ind2 = sub2ind(M.vnFy, np.c_[ii, jj + posFy]) + M.nFx

            IND = np.r_[ind1, ind2].flatten()

            PXX = sp.csr_matrix(
                (np.ones(2 * M.nC), (range(2 * M.nC), IND)), shape=(2 * M.nC, M.nF)
            )

            if M._meshType == "Curv":
                I2x2 = inverse_2x2_block_diagonal(
                    get_subarray(fN1[0], [i + posFx, j]),
                    get_subarray(fN1[1], [i + posFx, j]),
                    get_subarray(fN2[0], [i, j + posFy]),
                    get_subarray(fN2[1], [i, j + posFy]),
                )
                PXX = I2x2 * PXX

            return PXX

        return Pxx

    def _getFacePxxx(M):
        """returns a function for creating projection matrices

        Mats takes you from faces a subset of all faces on only the
        faces that you ask for.

        These are centered around a single nodes.
        """

        i, j, k = (
            np.arange(M.shape_cells[0]),
            np.arange(M.shape_cells[1]),
            np.arange(M.shape_cells[2]),
        )

        iijjkk = ndgrid(i, j, k)
        ii, jj, kk = iijjkk[:, 0], iijjkk[:, 1], iijjkk[:, 2]

        if M._meshType == "Curv":
            fN1 = M.reshape(M.face_normals, "F", "Fx", "M")
            fN2 = M.reshape(M.face_normals, "F", "Fy", "M")
            fN3 = M.reshape(M.face_normals, "F", "Fz", "M")

        def Pxxx(xFace, yFace, zFace):
            """
            xFace is 'fXp' or 'fXm'
            yFace is 'fYp' or 'fYm'
            zFace is 'fZp' or 'fZm'
            """

            # no  | node        | f1        | f2        | f3
            # 000 | i  ,j  ,k   | i  , j, k | i, j  , k | i, j, k
            # 100 | i+1,j  ,k   | i+1, j, k | i, j  , k | i, j, k
            # 010 | i  ,j+1,k   | i  , j, k | i, j+1, k | i, j, k
            # 110 | i+1,j+1,k   | i+1, j, k | i, j+1, k | i, j, k
            # 001 | i  ,j  ,k+1 | i  , j, k | i, j  , k | i, j, k+1
            # 101 | i+1,j  ,k+1 | i+1, j, k | i, j  , k | i, j, k+1
            # 011 | i  ,j+1,k+1 | i  , j, k | i, j+1, k | i, j, k+1
            # 111 | i+1,j+1,k+1 | i+1, j, k | i, j+1, k | i, j, k+1

            posX = 0 if xFace == "fXm" else 1
            posY = 0 if yFace == "fYm" else 1
            posZ = 0 if zFace == "fZm" else 1

            ind1 = sub2ind(M.vnFx, np.c_[ii + posX, jj, kk])
            ind2 = sub2ind(M.vnFy, np.c_[ii, jj + posY, kk]) + M.nFx
            ind3 = sub2ind(M.vnFz, np.c_[ii, jj, kk + posZ]) + M.nFx + M.nFy

            IND = np.r_[ind1, ind2, ind3].flatten()

            PXXX = sp.coo_matrix(
                (np.ones(3 * M.nC), (range(3 * M.nC), IND)), shape=(3 * M.nC, M.nF)
            ).tocsr()

            if M._meshType == "Curv":
                I3x3 = inverse_3x3_block_diagonal(
                    get_subarray(fN1[0], [i + posX, j, k]),
                    get_subarray(fN1[1], [i + posX, j, k]),
                    get_subarray(fN1[2], [i + posX, j, k]),
                    get_subarray(fN2[0], [i, j + posY, k]),
                    get_subarray(fN2[1], [i, j + posY, k]),
                    get_subarray(fN2[2], [i, j + posY, k]),
                    get_subarray(fN3[0], [i, j, k + posZ]),
                    get_subarray(fN3[1], [i, j, k + posZ]),
                    get_subarray(fN3[2], [i, j, k + posZ]),
                )
                PXXX = I3x3 * PXXX

            return PXXX

        return Pxxx

    def _getEdgePx(M):
        """Returns a function for creating projection matrices"""

        def Px(xEdge):
            if xEdge != "eX0":
                raise TypeError("xEdge = {0!s}, not eX0".format(xEdge))
            return sp.identity(M.nC)

        return Px

    def _getEdgePxx(M):
        i, j = np.arange(M.shape_cells[0]), np.arange(M.shape_cells[1])

        iijj = ndgrid(i, j)
        ii, jj = iijj[:, 0], iijj[:, 1]

        if M._meshType == "Curv":
            eT1 = M.reshape(M.edge_tangents, "E", "Ex", "M")
            eT2 = M.reshape(M.edge_tangents, "E", "Ey", "M")

        def Pxx(xEdge, yEdge):
            """
            no | node      | e1      | e2
            00 | i  ,j     | i  ,j   | i  ,j
            10 | i+1,j     | i  ,j   | i+1,j
            01 | i  ,j+1   | i  ,j+1 | i  ,j
            11 | i+1,j+1   | i  ,j+1 | i+1,j
            """
            posX = 0 if xEdge == "eX0" else 1
            posY = 0 if yEdge == "eY0" else 1

            ind1 = sub2ind(M.vnEx, np.c_[ii, jj + posX])
            ind2 = sub2ind(M.vnEy, np.c_[ii + posY, jj]) + M.nEx

            IND = np.r_[ind1, ind2].flatten()

            PXX = sp.coo_matrix(
                (np.ones(2 * M.nC), (range(2 * M.nC), IND)), shape=(2 * M.nC, M.nE)
            ).tocsr()

            if M._meshType == "Curv":
                I2x2 = inverse_2x2_block_diagonal(
                    get_subarray(eT1[0], [i, j + posX]),
                    get_subarray(eT1[1], [i, j + posX]),
                    get_subarray(eT2[0], [i + posY, j]),
                    get_subarray(eT2[1], [i + posY, j]),
                )
                PXX = I2x2 * PXX

            return PXX

        return Pxx

    def _getEdgePxxx(M):
        i, j, k = (
            np.arange(M.shape_cells[0]),
            np.arange(M.shape_cells[1]),
            np.arange(M.shape_cells[2]),
        )

        iijjkk = ndgrid(i, j, k)
        ii, jj, kk = iijjkk[:, 0], iijjkk[:, 1], iijjkk[:, 2]

        if M._meshType == "Curv":
            eT1 = M.reshape(M.edge_tangents, "E", "Ex", "M")
            eT2 = M.reshape(M.edge_tangents, "E", "Ey", "M")
            eT3 = M.reshape(M.edge_tangents, "E", "Ez", "M")

        def Pxxx(xEdge, yEdge, zEdge):
            """
            no  | node        | e1          | e2          | e3
            000 | i  ,j  ,k   | i  ,j  ,k   | i  ,j  ,k   | i  ,j  ,k
            100 | i+1,j  ,k   | i  ,j  ,k   | i+1,j  ,k   | i+1,j  ,k
            010 | i  ,j+1,k   | i  ,j+1,k   | i  ,j  ,k   | i  ,j+1,k
            110 | i+1,j+1,k   | i  ,j+1,k   | i+1,j  ,k   | i+1,j+1,k
            001 | i  ,j  ,k+1 | i  ,j  ,k+1 | i  ,j  ,k+1 | i  ,j  ,k
            101 | i+1,j  ,k+1 | i  ,j  ,k+1 | i+1,j  ,k+1 | i+1,j  ,k
            011 | i  ,j+1,k+1 | i  ,j+1,k+1 | i  ,j  ,k+1 | i  ,j+1,k
            111 | i+1,j+1,k+1 | i  ,j+1,k+1 | i+1,j  ,k+1 | i+1,j+1,k
            """

            posX = (
                [0, 0]
                if xEdge == "eX0"
                else [1, 0]
                if xEdge == "eX1"
                else [0, 1]
                if xEdge == "eX2"
                else [1, 1]
            )
            posY = (
                [0, 0]
                if yEdge == "eY0"
                else [1, 0]
                if yEdge == "eY1"
                else [0, 1]
                if yEdge == "eY2"
                else [1, 1]
            )
            posZ = (
                [0, 0]
                if zEdge == "eZ0"
                else [1, 0]
                if zEdge == "eZ1"
                else [0, 1]
                if zEdge == "eZ2"
                else [1, 1]
            )

            ind1 = sub2ind(M.vnEx, np.c_[ii, jj + posX[0], kk + posX[1]])
            ind2 = sub2ind(M.vnEy, np.c_[ii + posY[0], jj, kk + posY[1]]) + M.nEx
            ind3 = (
                sub2ind(M.vnEz, np.c_[ii + posZ[0], jj + posZ[1], kk]) + M.nEx + M.nEy
            )

            IND = np.r_[ind1, ind2, ind3].flatten()

            PXXX = sp.coo_matrix(
                (np.ones(3 * M.nC), (range(3 * M.nC), IND)), shape=(3 * M.nC, M.nE)
            ).tocsr()

            if M._meshType == "Curv":
                I3x3 = inverse_3x3_block_diagonal(
                    get_subarray(eT1[0], [i, j + posX[0], k + posX[1]]),
                    get_subarray(eT1[1], [i, j + posX[0], k + posX[1]]),
                    get_subarray(eT1[2], [i, j + posX[0], k + posX[1]]),
                    get_subarray(eT2[0], [i + posY[0], j, k + posY[1]]),
                    get_subarray(eT2[1], [i + posY[0], j, k + posY[1]]),
                    get_subarray(eT2[2], [i + posY[0], j, k + posY[1]]),
                    get_subarray(eT3[0], [i + posZ[0], j + posZ[1], k]),
                    get_subarray(eT3[1], [i + posZ[0], j + posZ[1], k]),
                    get_subarray(eT3[2], [i + posZ[0], j + posZ[1], k]),
                )
                PXXX = I3x3 * PXXX

            return PXXX

        return Pxxx

    # DEPRECATED
    getFaceInnerProduct = deprecate_method(
        "get_face_inner_product", "getFaceInnerProduct", removal_version="1.0.0", future_warn=False
    )
    getEdgeInnerProduct = deprecate_method(
        "get_edge_inner_product", "getEdgeInnerProduct", removal_version="1.0.0", future_warn=False
    )
    getFaceInnerProductDeriv = deprecate_method(
        "get_face_inner_product_deriv",
        "getFaceInnerProductDeriv",
        removal_version="1.0.0",
        future_warn=False
    )
    getEdgeInnerProductDeriv = deprecate_method(
        "get_edge_inner_product_deriv",
        "getEdgeInnerProductDeriv",
        removal_version="1.0.0",
        future_warn=False
    )



#      ___          ___       ___          ___          ___          ___
#     /\  \        /\  \     /\  \        /\  \        /\  \        /\  \
#    /::\  \      /::\  \    \:\  \      /::\  \      /::\  \      /::\  \
#   /:/\:\  \    /:/\:\  \    \:\  \    /:/\:\  \    /:/\:\  \    /:/\:\  \
#  /:/  \:\  \  /:/  \:\  \   /::\  \  /::\~\:\  \  /::\~\:\  \  /::\~\:\  \
# /:/__/ \:\__\/:/__/ \:\__\ /:/\:\__\/:/\:\ \:\__\/:/\:\ \:\__\/:/\:\ \:\__\
# \:\  \ /:/  /\:\  \  \/__//:/  \/__/\/_|::\/:/  /\:\~\:\ \/__/\:\~\:\ \/__/
#  \:\  /:/  /  \:\  \     /:/  /        |:|::/  /  \:\ \:\__\   \:\ \:\__\
#   \:\/:/  /    \:\  \    \/__/         |:|\/__/    \:\ \/__/    \:\ \/__/
#    \::/  /      \:\__\                 |:|  |       \:\__\       \:\__\
#     \/__/        \/__/                  \|__|        \/__/        \/__/
#
#
#
#                      .----------------.----------------.
#                     /|               /|               /|
#                    / |              / |              / |
#                   /  |      6      /  |     7       /  |
#                  /   |            /   |            /   |
#                 .----------------.----+-----------.    |
#                /|    . ---------/|----.----------/|----.
#               / |   /|         / |   /|         / |   /|
#              /  |  / |  4     /  |  / |   5    /  |  / |
#             /   | /  |       /   | /  |       /   | /  |
#            . -------------- .----------------.    |/   |
#            |    . ---+------|----.----+------|----.    |
#            |   /|    .______|___/|____.______|___/|____.
#            |  / |   /    2  |  / |   /     3 |  / |   /
#            | /  |  /        | /  |  /        | /  |  /
#            . ---+---------- . ---+---------- .    | /
#            |    |/          |    |/          |    |/             z
#            |    . ----------|----.-----------|----.              ^   y
#            |   /      0     |   /       1    |   /               |  /
#            |  /             |  /             |  /                | /
#            | /              | /              | /                 o----> x
#            . -------------- . -------------- .
#
#
# Face Refinement:
#
#      2_______________3                    _______________
#      |               |                   |       |       |
#   ^  |               |                   |   2   |   3   |
#   |  |               |                   |       |       |
#   |  |       x       |        --->       |-------+-------|
#   t1 |               |                   |       |       |
#      |               |                   |   0   |   1   |
#      |_______________|                   |_______|_______|
#      0      t0-->    1
#
#
# Face and Edge naming conventions:
#
#                      fZp
#                       |
#                 6 ------eX3------ 7
#                /|     |         / |
#               /eZ2    .        / eZ3
#             eY2 |        fYp eY3  |
#             /   |            / fXp|
#            4 ------eX2----- 5     |
#            |fXm 2 -----eX1--|---- 3          z
#           eZ0  /            |  eY1           ^   y
#            | eY0   .  fYm  eZ1 /             |  /
#            | /     |        | /              | /
#            0 ------eX0------1                o----> x
#                    |
#                   fZm
#
#
#            fX                                  fY
#      2___________3                       2___________3
#      |     e1    |                       |     e1    |
#      |           |                       |           |
#   e0 |     x     | e2      z          e0 |     x     | e2      z
#      |           |         ^             |           |         ^
#      |___________|         |___> y       |___________|         |___> x
#      0    e3     1                       0    e3     1
#           fZ
#      2___________3
#      |     e1    |
#      |           |
#   e0 |     x     | e2      y
#      |           |         ^
#      |___________|         |___> x
#      0    e3     1

from discretize.base import BaseTensorMesh
from discretize.operators import InnerProducts, DiffOperators
from discretize.mixins import InterfaceMixins, TreeMeshIO
from discretize.utils import as_array_n_by_dim
from discretize._extensions.tree_ext import _TreeMesh, TreeCell
import numpy as np
import scipy.sparse as sp
import warnings
from discretize.utils.code_utils import deprecate_property


class TreeMesh(
    _TreeMesh, BaseTensorMesh, InnerProducts, DiffOperators, TreeMeshIO, InterfaceMixins
):
    """Class for QuadTree (2D) and OcTree (3D) meshes.

    Tree meshes are numerical grids where the dimensions of each cell are powers of 2
    larger than some base cell dimension. Unlike the :class:`~discretize.TensorMesh`
    class, gridded locations and numerical operators for instances of ``TreeMesh``
    cannot be simply constructed using tensor products. Furthermore, each cell
    is an instance of ``TreeMesh`` is an instance of the
    :class:`~discretize.tree_mesh.TreeCell` .

    Parameters
    ----------
    h : (dim) iterable of int, numpy.ndarray, or tuple
        Defines the cell widths of the *underlying tensor mesh* along each axis. The
        length of the iterable object is equal to the dimension of the mesh (2 or 3).
        For a 3D mesh, the list would have the form *[hx, hy, hz]*. The number of cells
        along each axis **must be a power of 2** .

        Along each axis, the user has 3 choices for defining the cells widths for the
        underlying tensor mesh:

        - :class:`int` -> A unit interval is equally discretized into `N` cells.
        - :class:`numpy.ndarray` -> The widths are explicity given for each cell
        - the widths are defined as a :class:`list` of :class:`tuple` of the form *(dh, nc, [npad])*
          where *dh* is the cell width, *nc* is the number of cells, and *npad* (optional)
          is a padding factor denoting exponential increase/decrease in the cell width
          for each cell; e.g. *[(2., 10, -1.3), (2., 50), (2., 10, 1.3)]*

    origin : (dim) iterable, default: 0
        Define the origin or 'anchor point' of the mesh; i.e. the bottom-left-frontmost
        corner. By default, the mesh is anchored such that its origin is at [0, 0, 0].

        For each dimension (x, y or z), The user may set the origin 2 ways:

        - a ``scalar`` which explicitly defines origin along that dimension.
        - **{'0', 'C', 'N'}** a :class:`str` specifying whether the zero coordinate along
          each axis is the first node location ('0'), in the center ('C') or the last
          node location ('N') (see Examples).

    Examples
    --------
    Here we generate a basic 2D tree mesh.

    >>> from discretize import TreeMesh
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt

    Define base mesh (domain and finest discretization),

    >>> dh = 5    # minimum cell width (base mesh cell width)
    >>> nbc = 64  # number of base mesh cells
    >>> h = dh * np.ones(nbc)
    >>> mesh = TreeMesh([h, h])

    Define corner points for a rectangular box, and subdived the mesh within the box
    to the maximum refinement level.

    >>> x0s = [120.0, 80.0]
    >>> x1s = [240.0, 160.0]
    >>> levels = [mesh.max_level]
    >>> mesh.refine_box(x0s, x1s, levels)

    >>> mesh.plot_grid()
    >>> plt.show()
    """

    _meshType = "TREE"
    _aliases = {
        **BaseTensorMesh._aliases,
        **DiffOperators._aliases,
        **{
            "ntN": "n_total_nodes",
            "ntEx": "n_total_edges_x",
            "ntEy": "n_total_edges_y",
            "ntEz": "n_total_edges_z",
            "ntE": "n_total_edges",
            "ntFx": "n_total_faces_x",
            "ntFy": "n_total_faces_y",
            "ntFz": "n_total_faces_z",
            "ntF": "n_total_faces",
            "nhN": "n_hanging_nodes",
            "nhEx": "n_hanging_edges_x",
            "nhEy": "n_hanging_edges_y",
            "nhEz": "n_hanging_edges_z",
            "nhE": "n_hanging_edges",
            "nhFx": "n_hanging_faces_x",
            "nhFy": "n_hanging_faces_y",
            "nhFz": "n_hanging_faces_z",
            "nhF": "n_hanging_faces",
            "gridhN": "hanging_nodes",
            "gridhFx": "hanging_faces_x",
            "gridhFy": "hanging_faces_y",
            "gridhFz": "hanging_faces_z",
            "gridhEx": "hanging_edges_x",
            "gridhEy": "hanging_edges_y",
            "gridhEz": "hanging_edges_z",
        },
    }
    _items = {"h", "origin", "cell_state"}

    # inheriting stuff from BaseTensorMesh that isn't defined in _QuadTree
    def __init__(self, h=None, origin=None, **kwargs):
        if "x0" in kwargs:
            origin = kwargs.pop("x0")
        super().__init__(h=h, origin=origin)

        cell_state = kwargs.pop("cell_state", None)
        cell_indexes = kwargs.pop("cell_indexes", None)
        cell_levels = kwargs.pop("cell_levels", None)
        if cell_state is None:
            if cell_indexes is not None and cell_levels is not None:
                cell_state = {}
                cell_state["indexes"] = cell_indexes
                cell_state["levels"] = cell_levels
        if cell_state is not None:
            indexes = cell_state["indexes"]
            levels = cell_state["levels"]
            self.__setstate__((indexes, levels))

    def __repr__(self):
        """Plain text representation."""
        mesh_name = "{0!s}TreeMesh".format(("Oc" if self.dim == 3 else "Quad"))

        top = "\n" + mesh_name + ": {0:2.2f}% filled\n\n".format(self.fill * 100)

        # Number of cells per level
        level_count = self._count_cells_per_index()
        non_zero_levels = np.nonzero(level_count)[0]
        cell_display = ["Level : Number of cells"]
        cell_display.append("-----------------------")
        for level in non_zero_levels:
            cell_display.append("{:^5} : {:^15}".format(level, level_count[level]))
        cell_display.append("-----------------------")
        cell_display.append("Total : {:^15}".format(self.nC))

        extent_display = ["            Mesh Extent       "]
        extent_display.append("        min     ,     max     ")
        extent_display.append("   ---------------------------")
        dim_label = {0: "x", 1: "y", 2: "z"}
        for dim in range(self.dim):
            n_vector = getattr(self, "nodes_" + dim_label[dim])
            extent_display.append(
                "{}: {:^13},{:^13}".format(dim_label[dim], n_vector[0], n_vector[-1])
            )

        for i, line in enumerate(extent_display):
            if i == len(cell_display):
                cell_display.append(" " * (len(cell_display[0]) - 3 - len(line)))
            cell_display[i] += 3 * " " + line

        h_display = ["     Cell Widths    "]
        h_display.append("    min   ,   max   ")
        h_display.append("-" * (len(h_display[0])))
        h_gridded = self.h_gridded
        mins = np.min(h_gridded, axis=0)
        maxs = np.max(h_gridded, axis=0)
        for dim in range(self.dim):
            h_display.append("{:^10}, {:^10}".format(mins[dim], maxs[dim]))

        for i, line in enumerate(h_display):
            if i == len(cell_display):
                cell_display.append(" " * len(cell_display[0]))
            cell_display[i] += 3 * " " + line

        return top + "\n".join(cell_display)

    def _repr_html_(self):
        """html representation"""
        mesh_name = "{0!s}TreeMesh".format(("Oc" if self.dim == 3 else "Quad"))
        level_count = self._count_cells_per_index()
        non_zero_levels = np.nonzero(level_count)[0]
        dim_label = {0: "x", 1: "y", 2: "z"}
        h_gridded = self.h_gridded
        mins = np.min(h_gridded, axis=0)
        maxs = np.max(h_gridded, axis=0)

        style = " style='padding: 5px 20px 5px 20px;'"
        # Cell level table:
        cel_tbl = "<table>\n"
        cel_tbl += "<tr>\n"
        cel_tbl += "<th" + style + ">Level</th>\n"
        cel_tbl += "<th" + style + ">Number of cells</th>\n"
        cel_tbl += "</tr>\n"
        for level in non_zero_levels:
            cel_tbl += "<tr>\n"
            cel_tbl += "<td" + style + ">{}</td>\n".format(level)
            cel_tbl += "<td" + style + ">{}</td>\n".format(level_count[level])
            cel_tbl += "</tr>\n"
        cel_tbl += "<tr>\n"
        cel_tbl += (
            "<td style='font-weight: bold; padding: 5px 20px 5px 20px;'> Total </td>\n"
        )
        cel_tbl += "<td" + style + "> {} </td>\n".format(self.nC)
        cel_tbl += "</tr>\n"
        cel_tbl += "</table>\n"

        det_tbl = "<table>\n"
        det_tbl += "<tr>\n"
        det_tbl += "<th></th>\n"
        det_tbl += "<th" + style + " colspan='2'>Mesh extent</th>\n"
        det_tbl += "<th" + style + " colspan='2'>Cell widths</th>\n"
        det_tbl += "</tr>\n"

        det_tbl += "<tr>\n"
        det_tbl += "<th></th>\n"
        det_tbl += "<th" + style + ">min</th>\n"
        det_tbl += "<th" + style + ">max</th>\n"
        det_tbl += "<th" + style + ">min</th>\n"
        det_tbl += "<th" + style + ">max</th>\n"
        det_tbl += "</tr>\n"
        for dim in range(self.dim):
            n_vector = getattr(self, "nodes_" + dim_label[dim])
            det_tbl += "<tr>\n"
            det_tbl += "<td" + style + ">{}</td>\n".format(dim_label[dim])
            det_tbl += "<td" + style + ">{}</td>\n".format(n_vector[0])
            det_tbl += "<td" + style + ">{}</td>\n".format(n_vector[-1])
            det_tbl += "<td" + style + ">{}</td>\n".format(mins[dim])
            det_tbl += "<td" + style + ">{}</td>\n".format(maxs[dim])
            det_tbl += "</tr>\n"
        det_tbl += "</table>\n"

        full_tbl = "<table>\n"
        full_tbl += "<tr>\n"
        full_tbl += "<td style='font-weight: bold; font-size: 1.2em; text-align: center;'>{}</td>\n".format(
            mesh_name
        )
        full_tbl += "<td style='font-size: 1.2em; text-align: center;' colspan='2'>{0:2.2f}% filled</td>\n".format(
            100 * self.fill
        )
        full_tbl += "</tr>\n"
        full_tbl += "<tr>\n"

        full_tbl += "<td>\n"
        full_tbl += cel_tbl
        full_tbl += "</td>\n"

        full_tbl += "<td>\n"
        full_tbl += det_tbl
        full_tbl += "</td>\n"

        full_tbl += "</tr>\n"
        full_tbl += "</table>\n"

        return full_tbl

    @BaseTensorMesh.origin.setter
    def origin(self, value):
        # first use the BaseTensorMesh to set the origin to handle "0, C, N"
        BaseTensorMesh.origin.fset(self, value)
        # then update the TreeMesh with the hidden value
        self._set_origin(self._origin)

    @property
    def vntF(self):
        """
        Vector number of total faces along each axis

        This property returns the total number of hanging and
        non-hanging faces along each axis direction. The returned
        quantity is a list of integers of the form [nFx,nFy,nFz].

        Returns
        -------
        list of int
            Vector number of total faces along each axis
        """
        return [self.ntFx, self.ntFy] + ([] if self.dim == 2 else [self.ntFz])

    @property
    def vntE(self):
        """
        Vector number of total edges along each axis

        This property returns the total number of hanging and
        non-hanging edges along each axis direction. The returned
        quantity is a list of integers of the form [nEx,nEy,nEz].

        Returns
        -------
        list of int
            Vector number of total edges along each axis
        """
        return [self.ntEx, self.ntEy] + ([] if self.dim == 2 else [self.ntEz])

    @property
    def stencil_cell_gradient(self):
        if getattr(self, "_stencil_cell_gradient", None) is None:

            self._stencil_cell_gradient = sp.vstack(
                [self.stencil_cell_gradient_x, self.stencil_cell_gradient_y]
            )
            if self.dim == 3:
                self._stencil_cell_gradient = sp.vstack(
                    [self._stencil_cell_gradient, self.stencil_cell_gradient_z]
                )

        return self._stencil_cell_gradient

    @property
    def cell_gradient(self):
        if getattr(self, "_cell_gradient", None) is None:

            i_s = self.face_boundary_indices

            ix = np.ones(self.nFx)
            ix[i_s[0]] = 0.0
            ix[i_s[1]] = 0.0
            Pafx = sp.diags(ix)

            iy = np.ones(self.nFy)
            iy[i_s[2]] = 0.0
            iy[i_s[3]] = 0.0
            Pafy = sp.diags(iy)

            MfI = self.get_face_inner_product(invMat=True)

            if self.dim == 2:
                Pi = sp.block_diag([Pafx, Pafy])

            elif self.dim == 3:
                iz = np.ones(self.nFz)
                iz[i_s[4]] = 0.0
                iz[i_s[5]] = 0.0
                Pafz = sp.diags(iz)
                Pi = sp.block_diag([Pafx, Pafy, Pafz])

            self._cell_gradient = (
                -Pi * MfI * self.face_divergence.T * sp.diags(self.cell_volumes)
            )

        return self._cell_gradient

    @property
    def cell_gradient_x(self):
        if getattr(self, "_cell_gradient_x", None) is None:

            nFx = self.nFx
            i_s = self.face_boundary_indices

            ix = np.ones(self.nFx)
            ix[i_s[0]] = 0.0
            ix[i_s[1]] = 0.0
            Pafx = sp.diags(ix)

            MfI = self.get_face_inner_product(invMat=True)
            MfIx = sp.diags(MfI.diagonal()[:nFx])

            self._cell_gradient_x = (
                -Pafx * MfIx * self.face_x_divergence.T * sp.diags(self.cell_volumes)
            )

        return self._cell_gradient_x

    @property
    def cell_gradient_y(self):
        if getattr(self, "_cell_gradient_y", None) is None:

            nFx = self.nFx
            nFy = self.nFy
            i_s = self.face_boundary_indices

            iy = np.ones(self.nFy)
            iy[i_s[2]] = 0.0
            iy[i_s[3]] = 0.0
            Pafy = sp.diags(iy)

            MfI = self.get_face_inner_product(invMat=True)
            MfIy = sp.diags(MfI.diagonal()[nFx : nFx + nFy])

            self._cell_gradient_y = (
                -Pafy * MfIy * self.face_y_divergence.T * sp.diags(self.cell_volumes)
            )

        return self._cell_gradient_y

    @property
    def cell_gradient_z(self):
        if self.dim == 2:
            raise TypeError("z derivative not defined in 2D")
        if getattr(self, "_cell_gradient_z", None) is None:

            nFx = self.nFx
            nFy = self.nFy
            i_s = self.face_boundary_indices

            iz = np.ones(self.nFz)
            iz[i_s[4]] = 0.0
            iz[i_s[5]] = 0.0
            Pafz = sp.diags(iz)

            MfI = self.get_face_inner_product(invMat=True)
            MfIz = sp.diags(MfI.diagonal()[nFx + nFy :])

            self._cell_gradient_z = (
                -Pafz * MfIz * self.face_z_divergence.T * sp.diags(self.cell_volumes)
            )

        return self._cell_gradient_z

    @property
    def face_x_divergence(self):
        if getattr(self, "_face_x_divergence", None) is None:
            self._face_x_divergence = self.face_divergence[:, : self.nFx]
        return self._face_x_divergence

    @property
    def face_y_divergence(self):
        if getattr(self, "_face_y_divergence", None) is None:
            self._face_y_divergence = self.face_divergence[
                :, self.nFx : self.nFx + self.nFy
            ]
        return self._face_y_divergence

    @property
    def face_z_divergence(self):
        if getattr(self, "_face_z_divergence", None) is None:
            self._face_z_divergence = self.face_divergence[:, self.nFx + self.nFy :]
        return self._face_z_divergence

    def point2index(self, locs):
        """Finds cells that contain the given points.
        Returns an array of index values of the cells that contain the given
        points

        Parameters
        ----------
        locs: (N, dim) array_like
            points to search for the location of

        Returns
        -------
        (N) array_like of int
            Cell indices that contain the points
        """
        locs = as_array_n_by_dim(locs, self.dim)
        inds = self._get_containing_cell_indexes(locs)
        return inds

    def cell_levels_by_index(self, indices):
        """Fast function to return a list of levels for the given cell indices

        Parameters
        ----------
        index: (N) array_like
            Cell indexes to query

        Returns
        -------
        (N) numpy.ndarray of int
            Levels for the cells.
        """

        return self._cell_levels_by_indexes(indices)

    def get_interpolation_matrix(
        self, locs, location_type="CC", zeros_outside=False, **kwargs
    ):
        """Produces interpolation matrix

        Parameters
        ----------
        loc : (N, dim) array_like
            Location of points to interpolate to

        location_type: str, optional
            What to interpolate

            location_type can be:

            - 'CC'    -> scalar field defined on cell centers
            - 'Ex'    -> x-component of field defined on edges
            - 'Ey'    -> y-component of field defined on edges
            - 'Ez'    -> z-component of field defined on edges
            - 'Fx'    -> x-component of field defined on faces
            - 'Fy'    -> y-component of field defined on faces
            - 'Fz'    -> z-component of field defined on faces
            - 'N'     -> scalar field defined on nodes

        Returns
        -------
        (N, n_loc_type) scipy.sparse.csr_matrix
            the interpolation matrix

        """
        if "locType" in kwargs:
            warnings.warn(
                "The locType keyword argument has been deprecated, please use location_type. "
                "This will be removed in discretize 1.0.0",
                DeprecationWarning,
            )
            location_type = kwargs["locType"]
        if "zerosOutside" in kwargs:
            warnings.warn(
                "The zerosOutside keyword argument has been deprecated, please use zeros_outside. "
                "This will be removed in discretize 1.0.0",
                DeprecationWarning,
            )
            zeros_outside = kwargs["zerosOutside"]
        locs = as_array_n_by_dim(locs, self.dim)
        if location_type not in ["N", "CC", "Ex", "Ey", "Ez", "Fx", "Fy", "Fz"]:
            raise Exception(
                "location_type must be one of N, CC, Ex, Ey, Ez, Fx, Fy, or Fz"
            )

        if self.dim == 2 and location_type in ["Ez", "Fz"]:
            raise Exception("Unable to interpolate from Z edges/face in 2D")

        locs = np.require(np.atleast_2d(locs), dtype=np.float64, requirements="C")

        if location_type == "N":
            Av = self._getNodeIntMat(locs, zeros_outside)
        elif location_type in ["Ex", "Ey", "Ez"]:
            Av = self._getEdgeIntMat(locs, zeros_outside, location_type[1])
        elif location_type in ["Fx", "Fy", "Fz"]:
            Av = self._getFaceIntMat(locs, zeros_outside, location_type[1])
        elif location_type in ["CC"]:
            Av = self._getCellIntMat(locs, zeros_outside)
        return Av

    @property
    def permute_cells(self):
        """Permutation matrix re-ordering of cells sorted by x, then y, then z

        Returns
        -------
        (n_cells, n_cells) scipy.sparse.csr_matrix
        """
        # TODO: cache these?
        P = np.lexsort(self.gridCC.T)  # sort by x, then y, then z
        return sp.identity(self.nC).tocsr()[P]

    @property
    def permute_faces(self):
        """Permutation matrix re-ordering of faces sorted by x, then y, then z

        Returns
        -------
        (n_faces, n_faces) scipy.sparse.csr_matrix
        """
        # TODO: cache these?
        Px = np.lexsort(self.gridFx.T)
        Py = np.lexsort(self.gridFy.T) + self.nFx
        if self.dim == 2:
            P = np.r_[Px, Py]
        else:
            Pz = np.lexsort(self.gridFz.T) + (self.nFx + self.nFy)
            P = np.r_[Px, Py, Pz]
        return sp.identity(self.nF).tocsr()[P]

    @property
    def permute_edges(self):
        """Permutation matrix re-ordering of edges sorted by x, then y, then z

        Returns
        -------
        (n_edges, n_edges) scipy.sparse.csr_matrix
        """
        # TODO: cache these?
        Px = np.lexsort(self.gridEx.T)
        Py = np.lexsort(self.gridEy.T) + self.nEx
        if self.dim == 2:
            P = np.r_[Px, Py]
        if self.dim == 3:
            Pz = np.lexsort(self.gridEz.T) + (self.nEx + self.nEy)
            P = np.r_[Px, Py, Pz]
        return sp.identity(self.nE).tocsr()[P]

    @property
    def cell_state(self):
        """ The current state of the cells on the mesh.

        This represents the x, y, z indices of the cells in the base tensor mesh, as
        well as their levels. It can be used to reconstruct the mesh.

        Returns
        -------
        dict
            dictionary with two entries:

            - ``"indexes"``: the indexes of the cells
            - ``"levels"``: the levels of the cells
        """
        indexes, levels = self.__getstate__()
        return {"indexes": indexes.tolist(), "levels": levels.tolist()}

    def validate(self):
        return self.finalized

    def equals(self, other):
        try:
            if self.finalized and other.finalized:
                return super().equals(other)
        except AttributeError:
            pass
        return False

    def __reduce__(self):
        return TreeMesh, (self.h, self.origin), self.__getstate__()

    cellGrad = deprecate_property("cell_gradient", "cellGrad", removal_version="1.0.0", future_warn=False)
    cellGradx = deprecate_property(
        "cell_gradient_x", "cellGradx", removal_version="1.0.0", future_warn=False
    )
    cellGrady = deprecate_property(
        "cell_gradient_y", "cellGrady", removal_version="1.0.0", future_warn=False
    )
    cellGradz = deprecate_property(
        "cell_gradient_z", "cellGradz", removal_version="1.0.0", future_warn=False
    )
    cellGradStencil = deprecate_property(
        "cell_gradient_stencil", "cellGradStencil", removal_version="1.0.0", future_warn=False
    )
    nodalGrad = deprecate_property(
        "nodal_gradient", "nodalGrad", removal_version="1.0.0", future_warn=False
    )
    nodalLaplacian = deprecate_property(
        "nodal_laplacian", "nodalLaplacian", removal_version="1.0.0", future_warn=False
    )
    faceDiv = deprecate_property("face_divergence", "faceDiv", removal_version="1.0.0", future_warn=False)
    faceDivx = deprecate_property(
        "face_x_divergence", "faceDivx", removal_version="1.0.0", future_warn=False
    )
    faceDivy = deprecate_property(
        "face_y_divergence", "faceDivy", removal_version="1.0.0", future_warn=False
    )
    faceDivz = deprecate_property(
        "face_z_divergence", "faceDivz", removal_version="1.0.0", future_warn=False
    )
    edgeCurl = deprecate_property("edge_curl", "edgeCurl", removal_version="1.0.0", future_warn=False)
    maxLevel = deprecate_property("max_used_level", "maxLevel", removal_version="1.0.0", future_warn=False)
    vol = deprecate_property("cell_volumes", "vol", removal_version="1.0.0", future_warn=False)
    areaFx = deprecate_property("face_x_areas", "areaFx", removal_version="1.0.0", future_warn=False)
    areaFy = deprecate_property("face_y_areas", "areaFy", removal_version="1.0.0", future_warn=False)
    areaFz = deprecate_property("face_z_areas", "areaFz", removal_version="1.0.0", future_warn=False)
    area = deprecate_property("face_areas", "area", removal_version="1.0.0", future_warn=False)
    edgeEx = deprecate_property("edge_x_lengths", "edgeEx", removal_version="1.0.0", future_warn=False)
    edgeEy = deprecate_property("edge_y_lengths", "edgeEy", removal_version="1.0.0", future_warn=False)
    edgeEz = deprecate_property("edge_z_lengths", "edgeEz", removal_version="1.0.0", future_warn=False)
    edge = deprecate_property("edge_lengths", "edge", removal_version="1.0.0", future_warn=False)
    permuteCC = deprecate_property(
        "permute_cells", "permuteCC", removal_version="1.0.0", future_warn=False
    )
    permuteF = deprecate_property("permute_faces", "permuteF", removal_version="1.0.0", future_warn=False)
    permuteE = deprecate_property("permute_edges", "permuteE", removal_version="1.0.0", future_warn=False)
    faceBoundaryInd = deprecate_property(
        "face_boundary_indices", "faceBoundaryInd", removal_version="1.0.0", future_warn=False
    )
    cellBoundaryInd = deprecate_property(
        "cell_boundary_indices", "cellBoundaryInd", removal_version="1.0.0", future_warn=False
    )
    _aveCC2FxStencil = deprecate_property(
        "average_cell_to_total_face_x", "_aveCC2FxStencil", removal_version="1.0.0", future_warn=False
    )
    _aveCC2FyStencil = deprecate_property(
        "average_cell_to_total_face_y", "_aveCC2FyStencil", removal_version="1.0.0", future_warn=False
    )
    _aveCC2FzStencil = deprecate_property(
        "average_cell_to_total_face_z", "_aveCC2FzStencil", removal_version="1.0.0", future_warn=False
    )
    _cellGradStencil = deprecate_property(
        "stencil_cell_gradient", "_cellGradStencil", removal_version="1.0.0", future_warn=False
    )
    _cellGradxStencil = deprecate_property(
        "stencil_cell_gradient_x", "_cellGradxStencil", removal_version="1.0.0", future_warn=False
    )
    _cellGradyStencil = deprecate_property(
        "stencil_cell_gradient_y", "_cellGradyStencil", removal_version="1.0.0", future_warn=False
    )
    _cellGradzStencil = deprecate_property(
        "stencil_cell_gradient_z", "_cellGradzStencil", removal_version="1.0.0", future_warn=False
    )



from discretize.utils.code_utils import deprecate_module

deprecate_module(
    "discretize.View", "discretize.mixins.mpl_mod", removal_version="1.0.0", future_warn=False
)
try:
    from discretize.mixins.mpl_mod import Slicer
except ImportError:
    pass
